@@ -10,6 +10,8 @@
 **
 *************************************************************************
 **
+** $Id: test_async.c,v 1.56 2009/03/28 18:56:14 drh Exp $
+**
 ** This file contains an example implementation of an asynchronous IO 
 ** backend for SQLite.
 **
@@ -25,7 +27,7 @@
 ** You lose the Durable property.  With the default I/O backend of SQLite,
 ** once a write completes, you know that the information you wrote is
 ** safely on disk.  With the asynchronous I/O, this is not the case.  If
-** your program crashes or if a power lose occurs after the database
+** your program crashes or if a power loss occurs after the database
 ** write but before the asynchronous write thread has completed, then the
 ** database change might never make it to disk and the next user of the
 ** database might not see your change.
@@ -108,6 +110,8 @@
 
 #ifndef SQLITE_AMALGAMATION
 # include "sqliteInt.h"
+# include <assert.h>
+# include <string.h>
 #endif
 #include <tcl.h>
 
@@ -115,7 +119,7 @@
 ** This test uses pthreads and hence only works on unix and with
 ** a threadsafe build of SQLite.
 */
-#if OS_UNIX && SQLITE_THREADSAFE
+#if SQLITE_OS_UNIX && SQLITE_THREADSAFE
 
 /*
 ** This demo uses pthreads.  If you do not have a pthreads implementation
@@ -158,10 +162,10 @@ static void asyncTrace(const char *zFormat, ...){
 **       protected by the async.queueMutex. As are the async.ioError and
 **       async.nFile variables.
 **
-**     * The async.aLock hash-table and all AsyncLock and AsyncFileLock
+**     * The async.pLock list and all AsyncLock and AsyncFileLock
 **       structures must be protected by the async.lockMutex mutex.
 **
-**     * The file handles from the underlying system are assumed not to 
+**     * The file handles from the underlying system are not assumed to 
 **       be thread safe.
 **
 **     * See the last two paragraphs under "The Writer Thread" for
@@ -272,11 +276,11 @@ static struct TestAsyncStaticData {
   pthread_cond_t emptySignal;  /* Notify when the write queue is empty */
   AsyncWrite *pQueueFirst;     /* Next write operation to be processed */
   AsyncWrite *pQueueLast;      /* Last write operation on the list */
-  Hash aLock;                  /* Files locked */
+  AsyncLock *pLock;            /* Linked list of all AsyncLock structures */
   volatile int ioDelay;             /* Extra delay between write operations */
   volatile int writerHaltWhenIdle;  /* Writer thread halts when queue empty */
   volatile int writerHaltNow;       /* Writer thread halts after next op */
-  int ioError;                 /* True if an IO error has occured */
+  int ioError;                 /* True if an IO error has occurred */
   int nFile;                   /* Number of open files (from sqlite pov) */
 } async = {
   PTHREAD_MUTEX_INITIALIZER,
@@ -348,7 +352,7 @@ static const char *azOpcodeName[] = {
 struct AsyncWrite {
   AsyncFileData *pFileData;    /* File to write data to or sync */
   int op;                      /* One of ASYNC_xxx etc. */
-  i64 iOffset;        /* See above */
+  sqlite_int64 iOffset;        /* See above */
   int nByte;          /* See above */
   char *zBuf;         /* Data to write to file (or NULL if op!=ASYNC_WRITE) */
   AsyncWrite *pNext;  /* Next write operation (to any file) */
@@ -374,9 +378,12 @@ struct AsyncWrite {
 ** the implementation of database locking used by this backend.
 */
 struct AsyncLock {
+  char *zFile;
+  int nFile;
   sqlite3_file *pFile;
   int eLock;
   AsyncFileLock *pList;
+  AsyncLock *pNext;           /* Next in linked list headed by async.pLock */
 };
 
 /*
@@ -410,8 +417,9 @@ struct AsyncFileData {
   int nName;                 /* Number of characters in zName */
   sqlite3_file *pBaseRead;   /* Read handle to the underlying Os file */
   sqlite3_file *pBaseWrite;  /* Write handle to the underlying Os file */
-  AsyncFileLock lock;
-  AsyncWrite close;
+  AsyncFileLock lock;        /* Lock state for this handle */
+  AsyncLock *pLock;          /* AsyncLock object for this file system entry */
+  AsyncWrite closeOp;        /* Preallocated close operation */
 };
 
 /*
@@ -475,7 +483,6 @@ static int async_mutex_lock(pthread_mutex_t *pMutex){
   assert(&(aHolder[2])==&asyncdebug.writerMutexHolder);
 
   assert( pthread_self()!=0 );
-
   for(iIdx=0; iIdx<3; iIdx++){
     if( pMutex==&aMutex[iIdx] ) break;
 
@@ -563,12 +570,31 @@ static int async_cond_wait(pthread_cond_t *pCond, pthread_mutex_t *pMutex){
   return rc;
 }
 
+/*
+** Assert that the mutex is held by the current thread.
+*/
+static void assert_mutex_is_held(pthread_mutex_t *pMutex){
+  int iIdx;
+  pthread_mutex_t *aMutex = (pthread_mutex_t *)(&async);
+  pthread_t *aHolder = (pthread_t *)(&asyncdebug);
+
+  for(iIdx=0; iIdx<3; iIdx++){
+    if( pMutex==&aMutex[iIdx] ) break;
+  }
+  assert(iIdx<3);
+  assert( aHolder[iIdx]==pthread_self() );
+}
+
 /* Call our async_XX wrappers instead of selected pthread_XX functions */
 #define pthread_mutex_lock    async_mutex_lock
 #define pthread_mutex_unlock  async_mutex_unlock
 #define pthread_mutex_trylock async_mutex_trylock
 #define pthread_cond_wait     async_cond_wait
 
+#else    /* if defined(NDEBUG) */
+
+#define assert_mutex_is_held(X)    /* A no-op when not debugging */
+
 #endif   /* !defined(NDEBUG) */
 
 /*
@@ -583,7 +609,9 @@ static int async_cond_wait(pthread_cond_t *pCond, pthread_mutex_t *pMutex){
 */
 static void addAsyncWrite(AsyncWrite *pWrite){
   /* We must hold the queue mutex in order to modify the queue pointers */
-  pthread_mutex_lock(&async.queueMutex);
+  if( pWrite->op!=ASYNC_UNLOCK ){
+    pthread_mutex_lock(&async.queueMutex);
+  }
 
   /* Add the record to the end of the write-op queue */
   assert( !pWrite->pNext );
@@ -602,7 +630,9 @@ static void addAsyncWrite(AsyncWrite *pWrite){
   }
 
   /* Drop the queue mutex */
-  pthread_mutex_unlock(&async.queueMutex);
+  if( pWrite->op!=ASYNC_UNLOCK ){
+    pthread_mutex_unlock(&async.queueMutex);
+  }
 
   /* The writer thread might have been idle because there was nothing
   ** on the write-op queue for it to do.  So wake it up. */
@@ -629,7 +659,7 @@ static void incrOpenFileCount(){
 static int addNewAsyncWrite(
   AsyncFileData *pFileData, 
   int op, 
-  i64 iOffset, 
+  sqlite3_int64 iOffset, 
   int nByte,
   const char *zByte
 ){
@@ -674,7 +704,7 @@ static int asyncClose(sqlite3_file *pFile){
   p->lock.eLock = 0;
   pthread_mutex_unlock(&async.lockMutex);
 
-  addAsyncWrite(&p->close);
+  addAsyncWrite(&p->closeOp);
   return SQLITE_OK;
 }
 
@@ -684,7 +714,12 @@ static int asyncClose(sqlite3_file *pFile){
 ** the global AsyncWrite list. Either SQLITE_OK or SQLITE_NOMEM may be
 ** returned.
 */
-static int asyncWrite(sqlite3_file *pFile, const void *pBuf, int amt, i64 iOff){
+static int asyncWrite(
+  sqlite3_file *pFile, 
+  const void *pBuf, 
+  int amt, 
+  sqlite3_int64 iOff
+){
   AsyncFileData *p = ((AsyncFile *)pFile)->pData;
   return addNewAsyncWrite(p, ASYNC_WRITE, iOff, amt, pBuf);
 }
@@ -696,10 +731,15 @@ static int asyncWrite(sqlite3_file *pFile, const void *pBuf, int amt, i64 iOff){
 **
 ** This method holds the mutex from start to finish.
 */
-static int asyncRead(sqlite3_file *pFile, void *zOut, int iAmt, i64 iOffset){
+static int asyncRead(
+  sqlite3_file *pFile, 
+  void *zOut, 
+  int iAmt, 
+  sqlite3_int64 iOffset
+){
   AsyncFileData *p = ((AsyncFile *)pFile)->pData;
   int rc = SQLITE_OK;
-  i64 filesize;
+  sqlite3_int64 filesize;
   int nRead;
   sqlite3_file *pBase = p->pBaseRead;
 
@@ -715,13 +755,13 @@ static int asyncRead(sqlite3_file *pFile, void *zOut, int iAmt, i64 iOffset){
   }
 
   if( pBase->pMethods ){
-    rc = sqlite3OsFileSize(pBase, &filesize);
+    rc = pBase->pMethods->xFileSize(pBase, &filesize);
     if( rc!=SQLITE_OK ){
       goto asyncread_out;
     }
     nRead = MIN(filesize - iOffset, iAmt);
     if( nRead>0 ){
-      rc = sqlite3OsRead(pBase, zOut, nRead, iOffset);
+      rc = pBase->pMethods->xRead(pBase, zOut, nRead, iOffset);
       ASYNC_TRACE(("READ %s %d bytes at %d\n", p->zName, nRead, iOffset));
     }
   }
@@ -731,7 +771,10 @@ static int asyncRead(sqlite3_file *pFile, void *zOut, int iAmt, i64 iOffset){
     char *zName = p->zName;
 
     for(pWrite=async.pQueueFirst; pWrite; pWrite = pWrite->pNext){
-      if( pWrite->op==ASYNC_WRITE && pWrite->pFileData->zName==zName ){
+      if( pWrite->op==ASYNC_WRITE && (
+        (pWrite->pFileData==p) ||
+        (zName && pWrite->pFileData->zName==zName)
+      )){
         int iBeginOut = (pWrite->iOffset-iOffset);
         int iBeginIn = -iBeginOut;
         int nCopy;
@@ -757,7 +800,7 @@ asyncread_out:
 ** Truncate the file to nByte bytes in length. This just adds an entry to 
 ** the write-op list, no IO actually takes place.
 */
-static int asyncTruncate(sqlite3_file *pFile, i64 nByte){
+static int asyncTruncate(sqlite3_file *pFile, sqlite3_int64 nByte){
   AsyncFileData *p = ((AsyncFile *)pFile)->pData;
   return addNewAsyncWrite(p, ASYNC_TRUNCATE, nByte, 0, 0);
 }
@@ -778,10 +821,10 @@ static int asyncSync(sqlite3_file *pFile, int flags){
 **
 ** This method holds the mutex from start to finish.
 */
-int asyncFileSize(sqlite3_file *pFile, i64 *piSize){
+int asyncFileSize(sqlite3_file *pFile, sqlite3_int64 *piSize){
   AsyncFileData *p = ((AsyncFile *)pFile)->pData;
   int rc = SQLITE_OK;
-  i64 s = 0;
+  sqlite3_int64 s = 0;
   sqlite3_file *pBase;
 
   pthread_mutex_lock(&async.queueMutex);
@@ -793,18 +836,24 @@ int asyncFileSize(sqlite3_file *pFile, i64 *piSize){
   */
   pBase = p->pBaseRead;
   if( pBase->pMethods ){
-    rc = sqlite3OsFileSize(pBase, &s);
+    rc = pBase->pMethods->xFileSize(pBase, &s);
   }
 
   if( rc==SQLITE_OK ){
     AsyncWrite *pWrite;
     for(pWrite=async.pQueueFirst; pWrite; pWrite = pWrite->pNext){
-      if( pWrite->op==ASYNC_DELETE && strcmp(p->zName, pWrite->zBuf)==0 ){
+      if( pWrite->op==ASYNC_DELETE 
+       && p->zName 
+       && strcmp(p->zName, pWrite->zBuf)==0 
+      ){
         s = 0;
-      }else if( pWrite->pFileData && pWrite->pFileData->zName==p->zName){
+      }else if( pWrite->pFileData && (
+          (pWrite->pFileData==p) 
+       || (p->zName && pWrite->pFileData->zName==p->zName) 
+      )){
         switch( pWrite->op ){
           case ASYNC_WRITE:
-            s = MAX(pWrite->iOffset + (i64)(pWrite->nByte), s);
+            s = MAX(pWrite->iOffset + (sqlite3_int64)(pWrite->nByte), s);
             break;
           case ASYNC_TRUNCATE:
             s = MIN(s, pWrite->iOffset);
@@ -836,13 +885,13 @@ static int getFileLock(AsyncLock *pLock){
     }
 
     if( eRequired>pLock->eLock ){
-      rc = sqlite3OsLock(pLock->pFile, eRequired);
+      rc = pLock->pFile->pMethods->xLock(pLock->pFile, eRequired);
       if( rc==SQLITE_OK ){
         pLock->eLock = eRequired;
       }
     }
     else if( eRequired<pLock->eLock && eRequired<=SQLITE_LOCK_SHARED ){
-      rc = sqlite3OsUnlock(pLock->pFile, eRequired);
+      rc = pLock->pFile->pMethods->xUnlock(pLock->pFile, eRequired);
       if( rc==SQLITE_OK ){
         pLock->eLock = eRequired;
       }
@@ -852,6 +901,19 @@ static int getFileLock(AsyncLock *pLock){
   return rc;
 }
 
+/*
+** Return the AsyncLock structure from the global async.pLock list 
+** associated with the file-system entry identified by path zName 
+** (a string of nName bytes). If no such structure exists, return 0.
+*/
+static AsyncLock *findLock(const char *zName, int nName){
+  AsyncLock *p = async.pLock;
+  while( p && (p->nFile!=nName || memcmp(p->zFile, zName, nName)) ){
+    p = p->pNext;
+  }
+  return p;
+}
+
 /*
 ** The following two methods - asyncLock() and asyncUnlock() - are used
 ** to obtain and release locks on database files opened with the
@@ -861,58 +923,63 @@ static int asyncLock(sqlite3_file *pFile, int eLock){
   int rc = SQLITE_OK;
   AsyncFileData *p = ((AsyncFile *)pFile)->pData;
 
-  pthread_mutex_lock(&async.lockMutex);
-  if( p->lock.eLock<eLock ){
-    AsyncLock *pLock;
-    AsyncFileLock *pIter;
-    pLock = (AsyncLock *)sqlite3HashFind(&async.aLock, p->zName, p->nName);
-    assert(pLock && pLock->pList);
-    for(pIter=pLock->pList; pIter; pIter=pIter->pNext){
-      if( pIter!=&p->lock && (
-        (eLock==SQLITE_LOCK_EXCLUSIVE && pIter->eLock>=SQLITE_LOCK_SHARED) ||
-        (eLock==SQLITE_LOCK_PENDING && pIter->eLock>=SQLITE_LOCK_RESERVED) ||
-        (eLock==SQLITE_LOCK_RESERVED && pIter->eLock>=SQLITE_LOCK_RESERVED) ||
-        (eLock==SQLITE_LOCK_SHARED && pIter->eLock>=SQLITE_LOCK_PENDING)
-      )){
-        rc = SQLITE_BUSY;
+  if( p->zName ){
+    pthread_mutex_lock(&async.lockMutex);
+    if( p->lock.eLock<eLock ){
+      AsyncLock *pLock = p->pLock;
+      AsyncFileLock *pIter;
+      assert(pLock && pLock->pList);
+      for(pIter=pLock->pList; pIter; pIter=pIter->pNext){
+        if( pIter!=&p->lock && (
+          (eLock==SQLITE_LOCK_EXCLUSIVE && pIter->eLock>=SQLITE_LOCK_SHARED) ||
+          (eLock==SQLITE_LOCK_PENDING && pIter->eLock>=SQLITE_LOCK_RESERVED) ||
+          (eLock==SQLITE_LOCK_RESERVED && pIter->eLock>=SQLITE_LOCK_RESERVED) ||
+          (eLock==SQLITE_LOCK_SHARED && pIter->eLock>=SQLITE_LOCK_PENDING)
+        )){
+          rc = SQLITE_BUSY;
+        }
+      }
+      if( rc==SQLITE_OK ){
+        p->lock.eLock = eLock;
+        p->lock.eAsyncLock = MAX(p->lock.eAsyncLock, eLock);
+      }
+      assert(p->lock.eAsyncLock>=p->lock.eLock);
+      if( rc==SQLITE_OK ){
+        rc = getFileLock(pLock);
       }
     }
-    if( rc==SQLITE_OK ){
-      p->lock.eLock = eLock;
-      p->lock.eAsyncLock = MAX(p->lock.eAsyncLock, eLock);
-    }
-    assert(p->lock.eAsyncLock>=p->lock.eLock);
-    if( rc==SQLITE_OK ){
-      rc = getFileLock(pLock);
-    }
+    pthread_mutex_unlock(&async.lockMutex);
   }
-  pthread_mutex_unlock(&async.lockMutex);
 
   ASYNC_TRACE(("LOCK %d (%s) rc=%d\n", eLock, p->zName, rc));
   return rc;
 }
 static int asyncUnlock(sqlite3_file *pFile, int eLock){
+  int rc = SQLITE_OK;
   AsyncFileData *p = ((AsyncFile *)pFile)->pData;
-  AsyncFileLock *pLock = &p->lock;
-  pthread_mutex_lock(&async.lockMutex);
-  pLock->eLock = MIN(pLock->eLock, eLock);
-  pthread_mutex_unlock(&async.lockMutex);
-  return addNewAsyncWrite(p, ASYNC_UNLOCK, 0, eLock, 0);
+  if( p->zName ){
+    AsyncFileLock *pLock = &p->lock;
+    pthread_mutex_lock(&async.queueMutex);
+    pthread_mutex_lock(&async.lockMutex);
+    pLock->eLock = MIN(pLock->eLock, eLock);
+    rc = addNewAsyncWrite(p, ASYNC_UNLOCK, 0, eLock, 0);
+    pthread_mutex_unlock(&async.lockMutex);
+    pthread_mutex_unlock(&async.queueMutex);
+  }
+  return rc;
 }
 
 /*
 ** This function is called when the pager layer first opens a database file
 ** and is checking for a hot-journal.
 */
-static int asyncCheckReservedLock(sqlite3_file *pFile){
+static int asyncCheckReservedLock(sqlite3_file *pFile, int *pResOut){
   int ret = 0;
   AsyncFileLock *pIter;
-  AsyncLock *pLock;
   AsyncFileData *p = ((AsyncFile *)pFile)->pData;
 
   pthread_mutex_lock(&async.lockMutex);
-  pLock = (AsyncLock *)sqlite3HashFind(&async.aLock, p->zName, p->nName);
-  for(pIter=pLock->pList; pIter; pIter=pIter->pNext){
+  for(pIter=p->pLock->pList; pIter; pIter=pIter->pNext){
     if( pIter->eLock>=SQLITE_LOCK_RESERVED ){
       ret = 1;
     }
@@ -920,11 +987,12 @@ static int asyncCheckReservedLock(sqlite3_file *pFile){
   pthread_mutex_unlock(&async.lockMutex);
 
   ASYNC_TRACE(("CHECK-LOCK %d (%s)\n", ret, p->zName));
-  return ret;
+  *pResOut = ret;
+  return SQLITE_OK;
 }
 
 /* 
-** This is a no-op, as the asynchronous backend does not support locking.
+** sqlite3_file_control() implementation.
 */
 static int asyncFileControl(sqlite3_file *id, int op, void *pArg){
   switch( op ){
@@ -951,33 +1019,52 @@ static int asyncDeviceCharacteristics(sqlite3_file *pFile){
 }
 
 static int unlinkAsyncFile(AsyncFileData *pData){
-  AsyncLock *pLock;
   AsyncFileLock **ppIter;
   int rc = SQLITE_OK;
 
-  pLock = sqlite3HashFind(&async.aLock, pData->zName, pData->nName);
-  for(ppIter=&pLock->pList; *ppIter; ppIter=&((*ppIter)->pNext)){
-    if( (*ppIter)==&pData->lock ){
-      *ppIter = pData->lock.pNext;
-      break;
+  if( pData->zName ){
+    AsyncLock *pLock = pData->pLock;
+    for(ppIter=&pLock->pList; *ppIter; ppIter=&((*ppIter)->pNext)){
+      if( (*ppIter)==&pData->lock ){
+        *ppIter = pData->lock.pNext;
+        break;
+      }
     }
-  }
-  if( !pLock->pList ){
-    if( pLock->pFile ){
-      sqlite3OsClose(pLock->pFile);
+    if( !pLock->pList ){
+      AsyncLock **pp;
+      if( pLock->pFile ){
+        pLock->pFile->pMethods->xClose(pLock->pFile);
+      }
+      for(pp=&async.pLock; *pp!=pLock; pp=&((*pp)->pNext));
+      *pp = pLock->pNext;
+      sqlite3_free(pLock);
+    }else{
+      rc = getFileLock(pLock);
     }
-    sqlite3_free(pLock);
-    sqlite3HashInsert(&async.aLock, pData->zName, pData->nName, 0);
-    if( !sqliteHashFirst(&async.aLock) ){
-      sqlite3HashClear(&async.aLock);
-    }
-  }else{
-    rc = getFileLock(pLock);
   }
 
   return rc;
 }
 
+/*
+** The parameter passed to this function is a copy of a 'flags' parameter
+** passed to this modules xOpen() method. This function returns true
+** if the file should be opened asynchronously, or false if it should
+** be opened immediately.
+**
+** If the file is to be opened asynchronously, then asyncOpen() will add
+** an entry to the event queue and the file will not actually be opened
+** until the event is processed. Otherwise, the file is opened directly
+** by the caller.
+*/
+static int doAsynchronousOpen(int flags){
+  return (flags&SQLITE_OPEN_CREATE) && (
+      (flags&SQLITE_OPEN_MAIN_JOURNAL) ||
+      (flags&SQLITE_OPEN_TEMP_JOURNAL) ||
+      (flags&SQLITE_OPEN_DELETEONCLOSE)
+  );
+}
+
 /*
 ** Open a file.
 */
@@ -1006,13 +1093,18 @@ static int asyncOpen(
 
   sqlite3_vfs *pVfs = (sqlite3_vfs *)pAsyncVfs->pAppData;
   AsyncFile *p = (AsyncFile *)pFile;
-  int nName = strlen(zName)+1;
+  int nName = 0;
   int rc = SQLITE_OK;
   int nByte;
   AsyncFileData *pData;
   AsyncLock *pLock = 0;
   char *z;
-  int isExclusive = (flags&SQLITE_OPEN_EXCLUSIVE);
+  int isAsyncOpen = doAsynchronousOpen(flags);
+
+  /* If zName is NULL, then the upper layer is requesting an anonymous file */
+  if( zName ){
+    nName = strlen(zName)+1;
+  }
 
   nByte = (
     sizeof(AsyncFileData) +        /* AsyncFileData structure */
@@ -1029,44 +1121,52 @@ static int asyncOpen(
   pData->pBaseRead = (sqlite3_file*)z;
   z += pVfs->szOsFile;
   pData->pBaseWrite = (sqlite3_file*)z;
-  z += pVfs->szOsFile;
-  pData->zName = z;
-  pData->nName = nName;
-  pData->close.pFileData = pData;
-  pData->close.op = ASYNC_CLOSE;
-  memcpy(pData->zName, zName, nName);
+  pData->closeOp.pFileData = pData;
+  pData->closeOp.op = ASYNC_CLOSE;
 
-  if( !isExclusive ){
-    rc = sqlite3OsOpen(pVfs, zName, pData->pBaseRead, flags, pOutFlags);
-    if( rc==SQLITE_OK && ((*pOutFlags)&SQLITE_OPEN_READWRITE) ){
-      rc = sqlite3OsOpen(pVfs, zName, pData->pBaseWrite, flags, 0);
+  if( zName ){
+    z += pVfs->szOsFile;
+    pData->zName = z;
+    pData->nName = nName;
+    memcpy(pData->zName, zName, nName);
+  }
+
+  if( !isAsyncOpen ){
+    int flagsout;
+    rc = pVfs->xOpen(pVfs, pData->zName, pData->pBaseRead, flags, &flagsout);
+    if( rc==SQLITE_OK && (flagsout&SQLITE_OPEN_READWRITE) ){
+      rc = pVfs->xOpen(pVfs, pData->zName, pData->pBaseWrite, flags, 0);
+    }
+    if( pOutFlags ){
+      *pOutFlags = flagsout;
     }
   }
 
   pthread_mutex_lock(&async.lockMutex);
 
-  if( rc==SQLITE_OK ){
-    pLock = sqlite3HashFind(&async.aLock, pData->zName, pData->nName);
+  if( zName && rc==SQLITE_OK ){
+    pLock = findLock(pData->zName, pData->nName);
     if( !pLock ){
-      pLock = sqlite3MallocZero(pVfs->szOsFile + sizeof(AsyncLock));
+      int nByte = pVfs->szOsFile + sizeof(AsyncLock) + pData->nName + 1; 
+      pLock = (AsyncLock *)sqlite3_malloc(nByte);
       if( pLock ){
-        AsyncLock *pDelete;
+        memset(pLock, 0, nByte);
 #ifdef ENABLE_FILE_LOCKING
         if( flags&SQLITE_OPEN_MAIN_DB ){
           pLock->pFile = (sqlite3_file *)&pLock[1];
-          rc = sqlite3OsOpen(pVfs, zName, pLock->pFile, flags, 0);
+          rc = pVfs->xOpen(pVfs, pData->zName, pLock->pFile, flags, 0);
           if( rc!=SQLITE_OK ){
             sqlite3_free(pLock);
             pLock = 0;
           }
         }
 #endif
-        pDelete = sqlite3HashInsert(
-          &async.aLock, pData->zName, pData->nName, (void *)pLock
-        );
-        if( pDelete ){
-          rc = SQLITE_NOMEM;
-          sqlite3_free(pLock);
+        if( pLock ){
+          pLock->nFile = pData->nName;
+          pLock->zFile = &((char *)(&pLock[1]))[pVfs->szOsFile];
+          memcpy(pLock->zFile, pData->zName, pLock->nFile);
+          pLock->pNext = async.pLock;
+          async.pLock = pLock;
         }
       }else{
         rc = SQLITE_NOMEM;
@@ -1075,21 +1175,24 @@ static int asyncOpen(
   }
 
   if( rc==SQLITE_OK ){
-    HashElem *pElem;
     p->pMethod = &async_methods;
     p->pData = pData;
 
     /* Link AsyncFileData.lock into the linked list of 
     ** AsyncFileLock structures for this file.
     */
-    pData->lock.pNext = pLock->pList;
-    pLock->pList = &pData->lock;
-
-    pElem = sqlite3HashFindElem(&async.aLock, pData->zName, pData->nName);
-    pData->zName = (char *)sqliteHashKey(pElem);
+    if( zName ){
+      pData->lock.pNext = pLock->pList;
+      pLock->pList = &pData->lock;
+      pData->zName = pLock->zFile;
+    }
   }else{
-    sqlite3OsClose(pData->pBaseRead);
-    sqlite3OsClose(pData->pBaseWrite);
+    if( pData->pBaseRead->pMethods ){
+      pData->pBaseRead->pMethods->xClose(pData->pBaseRead);
+    }
+    if( pData->pBaseWrite->pMethods ){
+      pData->pBaseWrite->pMethods->xClose(pData->pBaseWrite);
+    }
     sqlite3_free(pData);
   }
 
@@ -1097,10 +1200,11 @@ static int asyncOpen(
 
   if( rc==SQLITE_OK ){
     incrOpenFileCount();
+    pData->pLock = pLock;
   }
 
-  if( rc==SQLITE_OK && isExclusive ){
-    rc = addNewAsyncWrite(pData, ASYNC_OPENEXCLUSIVE, (i64)flags, 0, 0);
+  if( rc==SQLITE_OK && isAsyncOpen ){
+    rc = addNewAsyncWrite(pData, ASYNC_OPENEXCLUSIVE, (sqlite3_int64)flags,0,0);
     if( rc==SQLITE_OK ){
       if( pOutFlags ) *pOutFlags = flags;
     }else{
@@ -1110,6 +1214,9 @@ static int asyncOpen(
       sqlite3_free(pData);
     }
   }
+  if( rc!=SQLITE_OK ){
+    p->pMethod = 0;
+  }
   return rc;
 }
 
@@ -1125,7 +1232,13 @@ static int asyncDelete(sqlite3_vfs *pAsyncVfs, const char *z, int syncDir){
 ** Implementation of sqlite3OsAccess. This method holds the mutex from
 ** start to finish.
 */
-static int asyncAccess(sqlite3_vfs *pAsyncVfs, const char *zName, int flags){
+static int asyncAccess(
+  sqlite3_vfs *pAsyncVfs, 
+  const char *zName, 
+  int flags,
+  int *pResOut
+){
+  int rc;
   int ret;
   AsyncWrite *p;
   sqlite3_vfs *pVfs = (sqlite3_vfs *)pAsyncVfs->pAppData;
@@ -1136,12 +1249,13 @@ static int asyncAccess(sqlite3_vfs *pAsyncVfs, const char *zName, int flags){
   );
 
   pthread_mutex_lock(&async.queueMutex);
-  ret = sqlite3OsAccess(pVfs, zName, flags);
-  if( flags==SQLITE_ACCESS_EXISTS ){
+  rc = pVfs->xAccess(pVfs, zName, flags, &ret);
+  if( rc==SQLITE_OK && flags==SQLITE_ACCESS_EXISTS ){
     for(p=async.pQueueFirst; p; p = p->pNext){
       if( p->op==ASYNC_DELETE && 0==strcmp(p->zBuf, zName) ){
         ret = 0;
       }else if( p->op==ASYNC_OPENEXCLUSIVE 
+             && p->pFileData->zName
              && 0==strcmp(p->pFileData->zName, zName) 
       ){
         ret = 1;
@@ -1154,12 +1268,8 @@ static int asyncAccess(sqlite3_vfs *pAsyncVfs, const char *zName, int flags){
     , zName, ret)
   );
   pthread_mutex_unlock(&async.queueMutex);
-  return ret;
-}
-
-static int asyncGetTempname(sqlite3_vfs *pAsyncVfs, int nBufOut, char *zBufOut){
-  sqlite3_vfs *pVfs = (sqlite3_vfs *)pAsyncVfs->pAppData;
-  return pVfs->xGetTempname(pVfs, nBufOut, zBufOut);
+  *pResOut = ret;
+  return rc;
 }
 
 /*
@@ -1173,47 +1283,34 @@ static int asyncFullPathname(
 ){
   int rc;
   sqlite3_vfs *pVfs = (sqlite3_vfs *)pAsyncVfs->pAppData;
-  rc = sqlite3OsFullPathname(pVfs, zPath, nPathOut, zPathOut);
+  rc = pVfs->xFullPathname(pVfs, zPath, nPathOut, zPathOut);
 
   /* Because of the way intra-process file locking works, this backend
   ** needs to return a canonical path. The following block assumes the
   ** file-system uses unix style paths. 
   */
   if( rc==SQLITE_OK ){
-    int iIn;
-    int iOut = 0;
-    int nPathOut = strlen(zPathOut);
-
-    for(iIn=0; iIn<nPathOut; iIn++){
-
-      /* Replace any occurences of "//" with "/" */
-      if( iIn<=(nPathOut-2) && zPathOut[iIn]=='/' && zPathOut[iIn+1]=='/'
-      ){
-        continue;
+    int i, j;
+    int n = nPathOut;
+    char *z = zPathOut;
+    while( n>1 && z[n-1]=='/' ){ n--; }
+    for(i=j=0; i<n; i++){
+      if( z[i]=='/' ){
+        if( z[i+1]=='/' ) continue;
+        if( z[i+1]=='.' && i+2<n && z[i+2]=='/' ){
+          i += 1;
+          continue;
+        }
+        if( z[i+1]=='.' && i+3<n && z[i+2]=='.' && z[i+3]=='/' ){
+          while( j>0 && z[j-1]!='/' ){ j--; }
+          if( j>0 ){ j--; }
+          i += 2;
+          continue;
+        }
       }
-
-      /* Replace any occurences of "/./" with "/" */
-      if( iIn<=(nPathOut-3) 
-       && zPathOut[iIn]=='/' && zPathOut[iIn+1]=='.' && zPathOut[iIn+2]=='/'
-      ){
-        iIn++;
-        continue;
-      }
-
-      /* Replace any occurences of "<path-component>/../" with "" */
-      if( iOut>0 && iIn<=(nPathOut-4) 
-       && zPathOut[iIn]=='/' && zPathOut[iIn+1]=='.' 
-       && zPathOut[iIn+2]=='.' && zPathOut[iIn+3]=='/'
-      ){
-        iIn += 3;
-        iOut--;
-        for( ; iOut>0 && zPathOut[iOut-1]!='/'; iOut--);
-        continue;
-      }
-
-      zPathOut[iOut++] = zPathOut[iIn];
+      z[j++] = z[i];
     }
-    zPathOut[iOut] = '\0';
+    z[j] = 0;
   }
 
   return rc;
@@ -1226,11 +1323,11 @@ static void asyncDlError(sqlite3_vfs *pAsyncVfs, int nByte, char *zErrMsg){
   sqlite3_vfs *pVfs = (sqlite3_vfs *)pAsyncVfs->pAppData;
   pVfs->xDlError(pVfs, nByte, zErrMsg);
 }
-static void *asyncDlSym(
+static void (*asyncDlSym(
   sqlite3_vfs *pAsyncVfs, 
   void *pHandle, 
   const char *zSymbol
-){
+))(void){
   sqlite3_vfs *pVfs = (sqlite3_vfs *)pAsyncVfs->pAppData;
   return pVfs->xDlSym(pVfs, pHandle, zSymbol);
 }
@@ -1261,7 +1358,6 @@ static sqlite3_vfs async_vfs = {
   asyncOpen,            /* xOpen */
   asyncDelete,          /* xDelete */
   asyncAccess,          /* xAccess */
-  asyncGetTempname,     /* xGetTempName */
   asyncFullPathname,    /* xFullPathname */
   asyncDlOpen,          /* xDlOpen */
   asyncDlError,         /* xDlError */
@@ -1282,14 +1378,9 @@ static sqlite3_vfs async_vfs = {
 static void asyncEnable(int enable){
   if( enable ){
     if( !async_vfs.pAppData ){
-      static int hashTableInit = 0;
       async_vfs.pAppData = (void *)sqlite3_vfs_find(0);
       async_vfs.mxPathname = ((sqlite3_vfs *)async_vfs.pAppData)->mxPathname;
       sqlite3_vfs_register(&async_vfs, 1);
-      if( !hashTableInit ){
-        sqlite3HashInit(&async.aLock, SQLITE_HASH_BINARY, 1);
-        hashTableInit = 1;
-      }
     }
   }else{
     if( async_vfs.pAppData ){
@@ -1394,27 +1485,31 @@ static void *asyncWriterThread(void *pIsStarted){
         assert( pBase );
         ASYNC_TRACE(("WRITE %s %d bytes at %d\n",
                 p->pFileData->zName, p->nByte, p->iOffset));
-        rc = sqlite3OsWrite(pBase, (void *)(p->zBuf), p->nByte, p->iOffset);
+        rc = pBase->pMethods->xWrite(pBase, (void *)(p->zBuf), p->nByte, p->iOffset);
         break;
 
       case ASYNC_SYNC:
         assert( pBase );
         ASYNC_TRACE(("SYNC %s\n", p->pFileData->zName));
-        rc = sqlite3OsSync(pBase, p->nByte);
+        rc = pBase->pMethods->xSync(pBase, p->nByte);
         break;
 
       case ASYNC_TRUNCATE:
         assert( pBase );
         ASYNC_TRACE(("TRUNCATE %s to %d bytes\n", 
                 p->pFileData->zName, p->iOffset));
-        rc = sqlite3OsTruncate(pBase, p->iOffset);
+        rc = pBase->pMethods->xTruncate(pBase, p->iOffset);
         break;
 
       case ASYNC_CLOSE: {
         AsyncFileData *pData = p->pFileData;
         ASYNC_TRACE(("CLOSE %s\n", p->pFileData->zName));
-        sqlite3OsClose(pData->pBaseWrite);
-        sqlite3OsClose(pData->pBaseRead);
+        if( pData->pBaseWrite->pMethods ){
+          pData->pBaseWrite->pMethods->xClose(pData->pBaseWrite);
+        }
+        if( pData->pBaseRead->pMethods ){
+          pData->pBaseRead->pMethods->xClose(pData->pBaseRead);
+        }
 
         /* Unlink AsyncFileData.lock from the linked list of AsyncFileLock 
         ** structures for this file. Obtain the async.lockMutex mutex 
@@ -1424,6 +1519,11 @@ static void *asyncWriterThread(void *pIsStarted){
         rc = unlinkAsyncFile(pData);
         pthread_mutex_unlock(&async.lockMutex);
 
+        if( !holdingMutex ){
+          pthread_mutex_lock(&async.queueMutex);
+          holdingMutex = 1;
+        }
+        assert_mutex_is_held(&async.queueMutex);
         async.pQueueFirst = p->pNext;
         sqlite3_free(pData);
         doNotFree = 1;
@@ -1431,23 +1531,60 @@ static void *asyncWriterThread(void *pIsStarted){
       }
 
       case ASYNC_UNLOCK: {
-        AsyncLock *pLock;
+        AsyncWrite *pIter;
         AsyncFileData *pData = p->pFileData;
         int eLock = p->nByte;
-        pthread_mutex_lock(&async.lockMutex);
-        pData->lock.eAsyncLock = MIN(
-            pData->lock.eAsyncLock, MAX(pData->lock.eLock, eLock)
-        );
-        assert(pData->lock.eAsyncLock>=pData->lock.eLock);
-        pLock = sqlite3HashFind(&async.aLock, pData->zName, pData->nName);
-        rc = getFileLock(pLock);
-        pthread_mutex_unlock(&async.lockMutex);
+
+        /* When a file is locked by SQLite using the async backend, it is 
+        ** locked within the 'real' file-system synchronously. When it is
+        ** unlocked, an ASYNC_UNLOCK event is added to the write-queue to
+        ** unlock the file asynchronously. The design of the async backend
+        ** requires that the 'real' file-system file be locked from the
+        ** time that SQLite first locks it (and probably reads from it)
+        ** until all asynchronous write events that were scheduled before
+        ** SQLite unlocked the file have been processed.
+        **
+        ** This is more complex if SQLite locks and unlocks the file multiple
+        ** times in quick succession. For example, if SQLite does: 
+        ** 
+        **   lock, write, unlock, lock, write, unlock
+        **
+        ** Each "lock" operation locks the file immediately. Each "write" 
+        ** and "unlock" operation adds an event to the event queue. If the
+        ** second "lock" operation is performed before the first "unlock"
+        ** operation has been processed asynchronously, then the first
+        ** "unlock" cannot be safely processed as is, since this would mean
+        ** the file was unlocked when the second "write" operation is
+        ** processed. To work around this, when processing an ASYNC_UNLOCK
+        ** operation, SQLite:
+        **
+        **   1) Unlocks the file to the minimum of the argument passed to
+        **      the xUnlock() call and the current lock from SQLite's point
+        **      of view, and
+        **
+        **   2) Only unlocks the file at all if this event is the last
+        **      ASYNC_UNLOCK event on this file in the write-queue.
+        */ 
+        assert( holdingMutex==1 );
+        assert( async.pQueueFirst==p );
+        for(pIter=async.pQueueFirst->pNext; pIter; pIter=pIter->pNext){
+          if( pIter->pFileData==pData && pIter->op==ASYNC_UNLOCK ) break;
+        }
+        if( !pIter ){
+          pthread_mutex_lock(&async.lockMutex);
+          pData->lock.eAsyncLock = MIN(
+              pData->lock.eAsyncLock, MAX(pData->lock.eLock, eLock)
+          );
+          assert(pData->lock.eAsyncLock>=pData->lock.eLock);
+          rc = getFileLock(pData->pLock);
+          pthread_mutex_unlock(&async.lockMutex);
+        }
         break;
       }
 
       case ASYNC_DELETE:
         ASYNC_TRACE(("DELETE %s\n", p->zBuf));
-        rc = sqlite3OsDelete(pVfs, p->zBuf, (int)p->iOffset);
+        rc = pVfs->xDelete(pVfs, p->zBuf, (int)p->iOffset);
         break;
 
       case ASYNC_OPENEXCLUSIVE: {
@@ -1455,7 +1592,7 @@ static void *asyncWriterThread(void *pIsStarted){
         AsyncFileData *pData = p->pFileData;
         ASYNC_TRACE(("OPEN %s flags=%d\n", p->zBuf, (int)p->iOffset));
         assert(pData->pBaseRead->pMethods==0 && pData->pBaseWrite->pMethods==0);
-        rc = sqlite3OsOpen(pVfs, pData->zName, pData->pBaseRead, flags, 0);
+        rc = pVfs->xOpen(pVfs, pData->zName, pData->pBaseRead, flags, 0);
         assert( holdingMutex==0 );
         pthread_mutex_lock(&async.queueMutex);
         holdingMutex = 1;
@@ -1478,17 +1615,18 @@ static void *asyncWriterThread(void *pIsStarted){
       async.pQueueLast = 0;
     }
     if( !doNotFree ){
+      assert_mutex_is_held(&async.queueMutex);
       async.pQueueFirst = p->pNext;
       sqlite3_free(p);
     }
     assert( holdingMutex );
 
-    /* An IO error has occured. We cannot report the error back to the
+    /* An IO error has occurred. We cannot report the error back to the
     ** connection that requested the I/O since the error happened 
     ** asynchronously.  The connection has already moved on.  There 
     ** really is nobody to report the error to.
     **
-    ** The file for which the error occured may have been a database or
+    ** The file for which the error occurred may have been a database or
     ** journal file. Regardless, none of the currently queued operations
     ** associated with the same database should now be performed. Nor should
     ** any subsequently requested IO on either a database or journal file 
@@ -1507,7 +1645,7 @@ static void *asyncWriterThread(void *pIsStarted){
 
     if( async.ioError && !async.pQueueFirst ){
       pthread_mutex_lock(&async.lockMutex);
-      if( 0==sqliteHashFirst(&async.aLock) ){
+      if( 0==async.pLock ){
         async.ioError = SQLITE_OK;
       }
       pthread_mutex_unlock(&async.lockMutex);
@@ -1520,7 +1658,7 @@ static void *asyncWriterThread(void *pIsStarted){
       pthread_mutex_unlock(&async.queueMutex);
       holdingMutex = 0;
       if( async.ioDelay>0 ){
-        sqlite3OsSleep(pVfs, async.ioDelay);
+        pVfs->xSleep(pVfs, async.ioDelay);
       }else{
         sched_yield();
       }
@@ -1692,7 +1830,7 @@ static int testAsyncWait(
 }
 
 
-#endif  /* OS_UNIX and SQLITE_THREADSAFE */
+#endif  /* SQLITE_OS_UNIX and SQLITE_THREADSAFE */
 
 /*
 ** This routine registers the custom TCL commands defined in this
@@ -1700,7 +1838,7 @@ static int testAsyncWait(
 ** of this module.
 */
 int Sqlitetestasync_Init(Tcl_Interp *interp){
-#if OS_UNIX && SQLITE_THREADSAFE
+#if SQLITE_OS_UNIX && SQLITE_THREADSAFE
   Tcl_CreateObjCommand(interp,"sqlite3async_enable",testAsyncEnable,0,0);
   Tcl_CreateObjCommand(interp,"sqlite3async_halt",testAsyncHalt,0,0);
   Tcl_CreateObjCommand(interp,"sqlite3async_delay",testAsyncDelay,0,0);
@@ -1708,6 +1846,6 @@ int Sqlitetestasync_Init(Tcl_Interp *interp){
   Tcl_CreateObjCommand(interp,"sqlite3async_wait",testAsyncWait,0,0);
   Tcl_LinkVar(interp, "sqlite3async_trace",
       (char*)&sqlite3async_trace, TCL_LINK_INT);
-#endif  /* OS_UNIX and SQLITE_THREADSAFE */
+#endif  /* SQLITE_OS_UNIX and SQLITE_THREADSAFE */
   return TCL_OK;
 }
