@@ -15,7 +15,7 @@
 ** correctly populates and syncs a journal file before writing to a
 ** corresponding database file.
 **
-** $Id: test_journal.c,v 1.11 2009/02/12 09:11:56 danielk1977 Exp $
+** $Id: test_journal.c,v 1.14 2009/03/28 17:21:52 danielk1977 Exp $
 */
 #if SQLITE_TEST          /* This file is used for testing only */
 
@@ -50,7 +50,7 @@
 **     c) The set of page numbers corresponding to free-list leaf pages.
 **     d) A check-sum for every page in the database file.
 **
-**   The start of a write-transaction is deemed to have occured when a 
+**   The start of a write-transaction is deemed to have occurred when a 
 **   28-byte journal header is written to byte offset 0 of the journal 
 **   file.
 **
@@ -206,6 +206,17 @@ struct JtGlobal {
 };
 static struct JtGlobal g = {0, 0};
 
+/*
+** Functions to obtain and relinquish a mutex to protect g.pList. The
+** STATIC_PRNG mutex is reused, purely for the sake of convenience.
+*/
+static void enterJtMutex(void){
+  sqlite3_mutex_enter(sqlite3_mutex_alloc(SQLITE_MUTEX_STATIC_PRNG));
+}
+static void leaveJtMutex(void){
+  sqlite3_mutex_leave(sqlite3_mutex_alloc(SQLITE_MUTEX_STATIC_PRNG));
+}
+
 extern int sqlite3_io_error_pending;
 static void stop_ioerr_simulation(int *piSave){
   *piSave = sqlite3_io_error_pending;
@@ -236,11 +247,12 @@ static int jtClose(sqlite3_file *pFile){
   jt_file *p = (jt_file *)pFile;
 
   closeTransaction(p);
+  enterJtMutex();
   if( p->zName ){
     for(pp=&g.pList; *pp!=p; pp=&(*pp)->pNext);
     *pp = p->pNext;
   }
-
+  leaveJtMutex();
   return sqlite3OsClose(p->pReal);
 }
 
@@ -257,7 +269,6 @@ static int jtRead(
   return sqlite3OsRead(p->pReal, zBuf, iAmt, iOfst);
 }
 
-
 /*
 ** Parameter zJournal is the name of a journal file that is currently 
 ** open. This function locates and returns the handle opened on the
@@ -274,6 +285,7 @@ static int jtRead(
 **/
 static jt_file *locateDatabaseHandle(const char *zJournal){
   jt_file *pMain = 0;
+  enterJtMutex();
   for(pMain=g.pList; pMain; pMain=pMain->pNext){
     int nName = strlen(zJournal) - strlen("-journal");
     if( (pMain->flags&SQLITE_OPEN_MAIN_DB)
@@ -284,6 +296,7 @@ static jt_file *locateDatabaseHandle(const char *zJournal){
       break;
     }
   }
+  leaveJtMutex();
   return pMain;
 }
 
@@ -656,6 +669,7 @@ static int jtOpen(
 ){
   int rc;
   jt_file *p = (jt_file *)pFile;
+  pFile->pMethods = 0;
   p->pReal = (sqlite3_file *)&p[1];
   p->pReal->pMethods = 0;
   rc = sqlite3OsOpen(g.pVfs, zName, p->pReal, flags, pOutFlags);
@@ -668,10 +682,12 @@ static int jtOpen(
     p->pNext = 0;
     p->pWritable = 0;
     p->aCksum = 0;
+    enterJtMutex();
     if( zName ){
       p->pNext = g.pList;
       g.pList = p;
     }
+    leaveJtMutex();
   }
   return rc;
 }
